var assert = require('assert');
var util = require('util');

var elliptic = require('../elliptic');
var BN = elliptic.bn;

// Just a common base for instance of check
function NIST() {
  this.p = null;
}
module.exports = NIST;

function w2b32(indices, n) {
  var src = [];
  for (var i = indices.length - 1; i >= 0; i--) {
    var index = indices[i];
    if (index === -1)
      src.push('0', '0');
    else
      src.push('w[' + index * 2 + ']', 'w[' + (index * 2 + 1) + ']');
  }
  src = 'return new BN([ ' + src.join(', ') + ' ]);';
  var fn = new Function('w', 'BN', src);
  fn.name = n;
  return fn;
}

function w2b64(indices, n) {
  var src = [];
  for (var i = indices.length - 1; i >= 0; i--) {
    var index = indices[i];
    if (index === -1)
      src.push('0', '0', '0', '0');
    else
      src.push('w[' + index * 4 + ']',
               'w[' + (index * 4 + 1) + ']',
               'w[' + (index * 4 + 2) + ']',
               'w[' + (index * 4 + 3) + ']');
  }
  src = 'return new BN([ ' + src.join(', ') + ' ]);';
  var fn = new Function('w', 'BN', src);
  fn.name = n;
  return fn;
}

var p192 = new BN('fffffffffffffffffffffffffffffffeffffffffffffffff', 16);
var p192_2 = p192.mul(p192);

var p256 = new BN(
    'ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff fffffffe fffffc2f',
    16);
var p256_2 = p256.mul(p256);

function P192() {
  NIST.call(this);
  this.p = p192;
  this.p2 = p192_2;
}
util.inherits(P192, NIST);
NIST.p192 = P192;

P192.prototype.t = w2b64([ 2, 1, 0 ], 't_192');
P192.prototype.s1 = w2b64([ -1, 3, 3 ], 's1_192');
P192.prototype.s2 = w2b64([ 4, 4, -1 ], 's2_192');
P192.prototype.s3 = w2b64([ 5, 5, 5 ], 's3_192');

P192.prototype.mod = function mod(a) {
  a.strip();
  assert(a.length <= 24);

  // a should be <= p^2
  if (a.cmp(this.p2) > 0)
    return a.mod(this.p);

  var cmp = a.cmp(this.p);
  if (cmp === 0)
    return new BN(0);
  else if (cmp < 0)
    return a.clone();

  var w = a.words.slice();
  for (var i = 0; i < 24; i++)
    w[i] |= 0;

  var t = this.t(w, BN);
  var s1 = this.s1(w, BN);
  var s2 = this.s2(w, BN);
  var s3 = this.s3(w, BN);

  var r = t.add(s1).add(s2).add(s3);
  return r.mod(this.p);
};

function P256() {
  NIST.call(this);
  this.p = p256;
  this.p2 = p256_2;
}
util.inherits(P256, NIST);
NIST.p256 = P256;
P256.prototype.t = w2b32([ 7, 6, 5, 4, 3, 2, 1, 0 ], 't_256');
P256.prototype.s1 = w2b32([ 15, 14, 13, 12, 11, -1, -1, -1 ], 's1_256');
P256.prototype.s2 = w2b32([ -1, 15, 14, 13, 12, -1, -1, -1 ], 's2_256');
P256.prototype.s3 = w2b32([ 15, 14, -1, -1, -1, 10, 9, 8 ], 's3_256');
P256.prototype.s4 = w2b32([ 8, 13, 15, 14, 13, 11, 10, 9 ], 's4_256');
P256.prototype.d1 = w2b32([ 10, 8, -1, -1, -1, 13, 12, 11 ], 'd1_256');
P256.prototype.d2 = w2b32([ 11, 9, -1, -1, 15, 14, 13, 12 ], 'd2_256');
P256.prototype.d3 = w2b32([ 12, -1, 10, 9, 8, 15, 14, 13 ], 'd3_256');
P256.prototype.d4 = w2b32([ 13, -1, 11, 10, 9, -1, 15, 14 ], 'd4_256');

P256.prototype.mod = function mod(a) {
  a.strip();
  assert(a.length <= 32);

  // a should be <= p^2
  if (a.cmp(this.p2) > 0)
    return a.mod(this.p);

  var cmp = a.cmp(this.p);
  if (cmp === 0)
    return new BN(0);
  else if (cmp < 0)
    return a.clone();

  var w = a.words.slice();
  for (var i = 0; i < 32; i++)
    w[i] |= 0;

  var t = this.t(w, BN);
  var s1 = this.s1(w, BN);
  var s2 = this.s2(w, BN);
  var s3 = this.s3(w, BN);
  var s4 = this.s4(w, BN);
  var d1 = this.d1(w, BN);
  var d2 = this.d2(w, BN);
  var d3 = this.d3(w, BN);
  var d4 = this.d4(w, BN);

  var r = t.add(s1.add(s2).shl(1)).add(s3).add(s4)
           .sub(d1).sub(d2).sub(d3).sub(d4);

  return r.mod(this.p);
};
